/*
** Module   : HandleChain.java
** Abstract : An abstract class which provides an  implementation for the most
**            common methods from CommonHandleChain.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CS  20130207 Created initial version.
** 002 CS  20130215 TYPE attribute must be saved as a string, not character.
** 003 CA  20130221 Extracted TYPE, unknown and readOnlyError into a separate superclass.
** 004 CA  20130529 Implemented proper resource chaining. Hold the first/last resource as context
**                  local info, per each chainable resource type.
** 005 CA  20130919 Resource cleanup must be done after it was deleted.
** 006 CA  20130927 FIRST-/LAST-resource maps are kept per resource type, not class (as a resource
**                  can have multiple implementations).
** 007 CA  20131024 Fixed getName(), to return a new character instance.
** 008 SVL 20131203 Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 009 CA  20131115 Changes related to properly instantiate dynamic resources.
** 010 CA  20131220 Fixed H006 - lastResource and firstResource must always use the resource type
**                  as key.  Changed c'tors so that it is possible to avoid updating the first and
**                  last references for a resource. 
*/

package com.goldencode.p2j.util;

import java.util.*;
import com.goldencode.p2j.security.*;

/**
 * This acts as a superclass which implements the most common methods of the
 * CommonHandleChain interface:
 * <table>
 * <tr>
 *    <th>attribute or base method</th>
 *    <th>API</th>
 * </tr>
 * <tr>
 *    <td>NAME</td>
 *    <td>{@link #name()}</td>
 * </tr>
 * <tr>
 *    <td>PREV-SIBLING</td>
 *    <td>{@link #getPrevSibling()}</td>
 * </tr>
 * <tr>
 *    <td>NEXT-SIBLING</td>
 *    <td>{@link #getNextSibling()}</td>
 * </tr>
 * <tr>
 *    <td>PRIVATE-DATA</td>
 *    <td>{@link #getPrivateData()}, {@link #setPrivateData(character)} and
 *        {@link #setPrivateData(String)}
 *    </td>
 * </tr>
 * </table>
 * <p>
 * The abstract class also provides setters for the implemented attributes 
 * which are not defined inside the interfaces to be used for initializing of
 * those attributes
 */
public abstract class HandleChain
extends HandleResource
implements CommonHandleChain,
           Deletable
{
   /** This context's chain-related data. */
   private static final ContextLocal<WorkArea> local =
   new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
   };
   
   /** Variable for holding the name attribute. */
   protected final character name = new character();
   
   /** Variable for holding the privateData attribute. */
   private final character privateData = new character();
   
   /** Variable for holding the PREV-SIBLING attribute. */
   private HandleChain prevSibling = null;
   
   /** Variable for holding the NEXT-SIBLING attribute. */
   private HandleChain nextSibling = null;

   /**
    * Default c'tor.
    */
   public HandleChain()
   {
      this(false, true);
   }
   
   
   /**
    * Create this dynamic or static resource.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public HandleChain(boolean dynamic)
   {
      this(dynamic, true);
   }
   
   /**
    * Create this dynamic or static resource.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    saveFirstLast
    *           Flag indicating if the first/last resource should be updated.
    */
   public HandleChain(boolean dynamic, boolean saveFirstLast)
   {
      super(dynamic);

      if (saveFirstLast)
      {
         WorkArea wa = local.get();
   
         String type = type();
   
         HandleChain last = wa.lastResource.get(type);
         if (last != null)
         {
            last.setNextSibling(this);
            this.setPrevSibling(last);
         }
         else
         {
            // this is the first resource in the chain
            wa.firstResource.put(type, this);
         }
         
         // always mark it as the last resource in the chain
         wa.lastResource.put(type, this);
      }
   }
   
   /**
    * Get the first resource in the chain, for the given class.
    * 
    * @param    type
    *           The resource type.
    *           
    * @return   The first resource in the chain.
    */
   static handle firstResource(String type)
   {
      return new handle(local.get().firstResource.get(type));
   }

   /**
    * Get the last resource in the chain, for the given class.
    * 
    * @param    type
    *           The resource type.
    *           
    * @return   The last resource in the chain.
    */
   static handle lastResource(String type)
   {
      return new handle(local.get().lastResource.get(type));
   }

   /**
    * Worker to be implemented by each resource.  Called by {@link #delete()}.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   protected abstract boolean resourceDelete();
   
   /**
    * Show a specific error, as this read-only <code>attribute</code> was used
    * on the right-side of an assignment.
    *  
    * @param    attribute
    *           The read-only attribute.
    */
   public void readOnlyError(String attribute)
   {
      handle.readOnlyError(new handle(this), attribute);
   }

   /**
    * Perform actual delete of an resource. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    */
   public void delete()
   {
      // call resource-specific delete
      if (!resourceDelete())
      {
         // if the resource-specific delete was not possible, exit
         return;
      }

      // the resource was deleted, clean up after it
      handle.removeResource(this);
      // remove from the widget pools too
      WidgetPool.removeResource(this);

      // check if the resource supports both of the NEXT-SIBLING and PREV-SIBLING. 
      // if not double chained, exit
      if (!(hasPrevSibling() && hasNextSibling()))
      {
         return;
      }

      // TODO: add support for only prev- or next- chains
      
      // else, remove it from the chain
      if (prevSibling != null)
      {
         prevSibling.nextSibling = nextSibling;
      }
      if (nextSibling != null)
      {
         nextSibling.prevSibling = prevSibling;
      }
      
      String type = type();
      WorkArea wa = local.get();
      HandleChain first = wa.firstResource.get(type);
      HandleChain last = wa.lastResource.get(type);
      
      if (this == first)
      {
         wa.firstResource.put(type, this.nextSibling);
      }
      if (this == last)
      {
         wa.lastResource.put(type, this.prevSibling);
      }

      this.nextSibling = null;
      this.prevSibling = null;
   }
   
   /**
    * Get the <code>name</code> attribute of handle.
    * 
    * @return   See above.
    */
   public character name()
   {
      if (!hasName())
      {
         invalidAttribute("NAME");

         return new character();
      }

      return new character(name);
   }
   
   /**
    * Sets the name of this handle.This method is outside interfaces and only 
    * meant for initializing this within the implementation.
    * 
    * @param name
    *        The handle type to be set.
    */
   public void name(character name)
   {
      if (hasNameReadOnly())
      {
         readOnlyError("NAME");
         
         return;
      }

      this.name.assign(name); 
   }
   
   /**
    * Sets the name of this handle.This method is outside interfaces and only 
    * meant for initializing this within the implementation.
    * 
    * @param name
    *        The handle type to be set.
    */
   public void name(String name)
   {
      name(new character(name));
   }

   /**
    * Get the <code>next-sibling</code> attribute of this handle.
    * 
    * @return   See above.
    */
   public handle getNextSibling()
   {
      if (!hasNextSibling())
      {
         invalidAttribute("PREV-SIBLING");
         
         return new handle();
      }

      return new handle(nextSibling);
   }
   
   /**
    * Sets the NEXT-SIBLING attribute of this handle.This method is outside 
    * interfaces and only meant for initializing this within the implementation.
    * 
    * @param    nextSibling
    *           The next sibling handle in the chain.
    */
   public void setNextSibling(handle nextSibling)
   {
      if (!hasNextSibling())
         return;

      setNextSibling((HandleChain) nextSibling.getResource());
   }

   /**
    * Sets the NEXT-SIBLING attribute of this handle.This method is outside 
    * interfaces and only meant for initializing this within the implementation.
    * 
    * @param    nextSibling
    *           The next sibling resource in the chain.
    */
   public void setNextSibling(HandleChain nextSibling)
   {
      if (!hasNextSibling())
         return;

      this.nextSibling = nextSibling;
   }

   /**
    * Get the <code>prev-sibling</code> attribute of this handle.
    * 
    * @return   See above.
    */
   public handle getPrevSibling()
   {
      if (!hasPrevSibling())
      {
         invalidAttribute("NEXT-SIBLING");
         
         return new handle();
      }

      return new handle(prevSibling);
   }
   
   /**
    * Sets the PREV-SIBLING attribute of this handle.This method is outside 
    * interfaces and only meant for initializing this within the implementation.
    * 
    * @param    prevSibling
    *           The previous sibling handle in the chain.
    */
   public void setPrevSibling(handle prevSibling)
   {
      if (!hasPrevSibling())
         return;
      
      setPrevSibling((HandleChain) prevSibling.getResource());
   }
   
   /**
    * Sets the PREV-SIBLING attribute of this handle.This method is outside interfaces and only
    * meant for initializing this within the implementation.
    * 
    * @param    prevSibling
    *           The previous sibling resource in the chain.
    */
   public void setPrevSibling(HandleChain prevSibling)
   {
      if (!hasPrevSibling())
         return;
      
      this.prevSibling = prevSibling;
   }

   /**
    * Get the <code>private-data</code> attribute of this handle.
    * 
    * @return   See above.
    */
   public character getPrivateData()
   {
      if (!hasPrivateData())
      {
         invalidAttribute("PRIVATE-DATA");
         
         return new character();
      }
      
      return new character(privateData);
   }

   /**
    * Set the <code>private-data</code> attribute of this handle.
    * 
    * @param    data
    *           The new private-data value. 
    */
   public void setPrivateData(character data)
   {
      if (!hasPrivateData())
      {
         readOnlyError("PRIVATE-DATA");
         
         return;
      }

      this.privateData.assign(data);
   }

   /**
    * Set the <code>private-data</code> attribute of this handle.
    * 
    * @param    data
    *           The new private-data value. 
    */
   public void setPrivateData(String data)
   {
      setPrivateData(new character(data));
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   <code>true</code> if the resource has the attribute.
    */
   protected boolean hasNextSibling()
   {
      return true;
   }
   
   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   <code>true</code> if the resource has the attribute.
    */
   protected boolean hasPrevSibling()
   {
      return true;
   }
   
   /**
    * Check if this resource supports the NAME attribute in read-only mode.
    * 
    * @return   <code>true</code> if the resource has the attribute in r/o mode.
    */
   protected boolean hasNameReadOnly()
   {
      return true;
   }
   
   /**
    * Check if this resource supports the NAME attribute.
    * 
    * @return   <code>true</code> if the resource has the attribute.
    */
   protected boolean hasName()
   {
      return true;
   }
   
   /**
    * Check if this resource supports the PRIVATE-DATA attribute.
    * 
    * @return   <code>true</code> if the resource has the attribute.
    */
   protected boolean hasPrivateData()
   {
      return true;
   }
   
   /** 
    * Stores global data relating to the state of the current context.
    */
   private static class WorkArea
   {
      /** The first connected resource in the chain, per resource type. */
      private Map<String, HandleChain> firstResource = new HashMap<String, HandleChain>();
      
      /** The last connected resource in the chain, per resource type. */
      private Map<String, HandleChain> lastResource = new HashMap<String, HandleChain>();
   }
}
